import cv2
import math
import numpy as np
import random
import torch

from PIL import Image, ImageDraw
from skimage import io
from torch.utils.data import Dataset
from torchvision import transforms
import copy
"""

Dataset Defintion

Image: Min Width = 163 and Min Height = 143

"""
import re
def find_replacement(m):
    return replacements[m.group(1)]

replacements = {
    'bigneuron': 'bigneuron.edu'
}



class CountingDataset(Dataset):
    
    def __init__(self, 
                 data_split,
                 split_type,
                 annotations, 
                 num_examples=3, 
                 example_type="Box",
                 transform=None, 
                 do_cache=False,src_dot_dir=None,crop=False,cropSize=384,return_dot=False):
        
        self.data_split = data_split
        self.split_type = split_type
        self.annotations = annotations
        self.num_examples = num_examples
        self.example_type = example_type
        self.transform = transform
        self.do_cache = do_cache
        self.cache = {}
        self.return_dot = return_dot
        if src_dot_dir is not None:
            self.dots = True
            self.src_dot_dir = src_dot_dir
        else:
            self.dots = False
        self.crop = crop
        self.cropSize = cropSize
    
    def __len__(self):
        return len(self.data_split)
    
    def __getitem__(self, idx):
        
        annotations = self.annotations[self.data_split[idx]]
        
        img_path, density_path = annotations["img_path"], annotations["density_path"]
        regex = r'({})'.format(r'|'.join(re.escape(w) for w in replacements))
        img_path1 = re.sub(regex, find_replacement, img_path)        
        regex = r'({})'.format(r'|'.join(re.escape(w) for w in replacements))
        density_path1 = re.sub(regex, find_replacement, density_path)
        
        image = self.__fetch_image__(img_path1)
        density = self.__fetch_density__(density_path1)
        if self.dots:
            name_dot = density_path.split("/")[-1]
            src_dot = self.__fetch_dots__(name_dot)
        # Ensuring density_map and image shape is same.
        if not (density.shape[0] == image.size[1] and density.shape[1] == image.size[0]):
            raise RuntimeError("Image: {} {} and GT_Density: {} {} shape must be same.".format(
                img_path, image.size, density_path, density.shape))
                    
            
        if self.example_type.lower() == "Box".lower():
            #sampled_examples_path = self.__sample__(annotations["box_examples_path"], self.num_examples, self.split_type)
            
            #examples = [self.__fetch_image__(p) for p in sampled_examples_path]
            try:
                boxes = self.__sample__(annotations["box_examples_coordinates"], self.num_examples, self.split_type)
            except:
                boxes = self.__sample__(annotations["box_examples_coordinates"], self.num_examples - 1, self.split_type)
            if len(boxes) == 1:
                Boxes = [0] + [int(item) for item in boxes[0][0] + boxes[0][-2]]
                Boxes = torch.Tensor(Boxes)
            else:
                Boxes = list()
                for box in boxes:
                    a = [0] + [int(item) for item in box[0] + box[-2]]
                    Boxes.append(a)
                Boxes = torch.Tensor(Boxes)
        else:
            raise RuntimeError("Invalid example type: " + self.example_type)
        
        if self.dots:
            sample = {
            'image': image, 
            'density': density,
            'boxes': Boxes,
            'path': img_path,
            'src_dot': src_dot, 
            }
        else:
            sample = {
            'image': image, 
            'density': density,
            'boxes': Boxes,
            'path': img_path}            

        if self.transform: sample = self.transform(sample)
        return sample
    
    def __sample__(self, a_list, n, split_type):
        
        if split_type.lower() == "train": 
            return random.sample(a_list, n)
        elif split_type.lower() == "val": 
            return a_list[:n]
        else:
            raise RuntimeError("Invalid split type: " + split_type)
    
    def __fetch_image__(self, img_path):
        
        #if img_path in self.cache: return self.cache[img_path]
        
        # Reading image from the file
        image = Image.open(img_path)
        image.load()
        
        #if self.do_cache: self.cache[img_path] = image
            
        return image
    
    def __fetch_density__(self, density_path):
        
        #if density_path in self.cache: return self.cache[density_path]
        
        # Reading Density Map from the file
        density = np.load(density_path).astype('float32')
    
        #if self.do_cache: self.cache[density_path] = density
            
        return density

    def __fetch_dots__(self, dot_name):
        
        #if density_path in self.cache: return self.cache[density_path]
        dot_path = self.src_dot_dir + '/' + dot_name
        # Reading Density Map from the file
        src_dot = np.load(dot_path).astype('float32')
        #dot_path = self.tgt_dot_dir + '/' + dot_name
        # Reading Density Map from the file
        #tgt_dot = np.load(dot_path).astype('float32')
        #dot_path = self.tgt_density_dir + '/' + dot_name
        # Reading Density Map from the file
        #tgt_density = np.load(dot_path).astype('float32')
        #if self.do_cache: self.cache[density_path] = density
            
        return src_dot
    
    def __getname__(self, idx):
        
        return self.data_split[idx]
    
    def __closest_point__(self, points, p):
        
        d = [math.sqrt((p[0] - q[0])**2 + (p[1] - q[1])**2) for q in points]
        i = np.argmin(d)
    
        return (points[i], d[i])
"""

Custom Transforms

"""

class ResizeImage(object):
    """
    Resize the image in a sample to the nearest multiple of 8.
    """
    
    def __init__(self, MAX_HW,crop=False,dots=False,return_dot=False):
        self.MAX_HW = MAX_HW
        #self.MIN_HW = MIN_HW
        self.crop = crop
        self.dots = dots
        self.return_dot = return_dot

    def __call__(self, sample):
        image, density,img_path,boxes = sample['image'], sample['density'],sample['path'],sample['boxes']
        if self.dots:
            src_dot = sample['src_dot']
        
        W, H = image.size
        W0, H0 = image.size
        """
        Ensuring image size does not cross MAX_HW. 
        New size will be computed by resizing max(H,W) to MAX_HW and keeping the aspect ratio same.
        Referred: https://discuss.pytorch.org/t/no-resize-in-torchvision-transforms/10549/2
        """
        if W > self.MAX_HW:
            #H = int(round(self.MAX_HW * (float(H)/max(H, W))))
            W = int(round(self.MAX_HW * (float(W)/max(H, W))))
            #W0 = int(round(self.MAX_HW * (float(W)/max(H, W))))       
        ratio = W/(H*1.)
        #if W < self.MIN_HW:
        #    H = H
        #    W = self.MIN_HW
        H = int(round(H))
        W = int(round(W))
        new_W, new_H = W - (W%8), H - (H%8)
        #print(new_W,new_H)
        """
        Image is resized to the above computed size.
        """
        resized_image = transforms.Resize((new_H, new_W))(image)
        """
        Density is resized to image size.
        """
        resized_density = cv2.resize(density, (new_W, new_H))
        orig_count = np.sum(density)
        new_count = np.sum(resized_density)
        
        if new_count > 0: resized_density = resized_density * (orig_count / new_count)
        
        M = boxes.shape[0]
        #print(M)
        #print(boxes)
        for j in range(0,M): # for each provided examplar bounding box
            x1,y1 = math.ceil(boxes[j,1]),math.ceil(boxes[j,2])
            x2, y2 = math.ceil(boxes[j,3]),math.ceil(boxes[j,4]) 
            y1 = y1 * (new_H/H0)
            x1 = x1 * (new_W/W0)
            y2 = y2 * (new_H/H0)
            x2 = x2 * (new_W/W0) 
            boxes[j,1], boxes[j,2], boxes[j,3], boxes[j,4] = int(x1), int(y1), int(x2), int(y2)
        #print('new',boxes)
        if self.dots:
            src_dotNew = np.float32(np.zeros(shape=(new_H, new_W)))
            #tgt_dotNew = np.float32(np.zeros(shape=(new_H, new_W)))
            indices_y,indices_x = np.nonzero(src_dot)
            indices_y = indices_y * (new_H/H0)
            indices_x = indices_x * (new_W/W0)
            points = list()
            for i in range(0,len(indices_x)):
                points.append([indices_y[i],indices_x[i]])            
            dots_int = np.array(points).astype(int)
            #print(H0,W0,new_H,new_W)
            #print(dots_int)
            src_dotNew[dots_int[:, 0], dots_int[:, 1]] = 1            
            #resized_tgt_density = cv2.resize(tgt_density, (new_W, new_H))
            #orig_count = np.sum(tgt_density)
            #new_count = np.sum(resized_tgt_density)

            #if new_count > 0: resized_tgt_density = resized_tgt_density * (orig_count / new_count)            
            
            return {
                'image': resized_image, 
                'density': resized_density,
                'boxes': boxes,#sample['boxes'],
                'path': img_path,
                'src_dot': src_dotNew,
                'points':points
            }
        else:
            return {
                'image': resized_image, 
                'density': resized_density,
                'boxes': boxes,#sample['boxes'],
                'path': img_path,
            }            
class ResizeExampleImage(object):
    """
    Resize the examples in a sample to max example size in the sample.
    """

    def __call__(self, sample):
        examples = sample['examples']

        example_size = (0, 0) # (H, W)
        for ex in examples:
            example_size = (max(ex.size[1], example_size[0]), max(ex.size[0], example_size[1]))
        resize_example = transforms.Resize(example_size)
        
        resized_examples = [resize_example(example) for example in examples]
                            
        return {
            'image': sample['image'], 
            'density': sample['density'],
            'examples': resized_examples,
        }

# Taken from: https://pytorch.org/tutorials/beginner/data_loading_tutorial.html#transforms

class ToTensor(object):
    """Convert ndarrays in sample to Tensors."""
    
    def __init__(self,dots=False):
        
        self.to_tensor = transforms.ToTensor()
        self.dots = dots
    def __call__(self, sample):
        image, density, boxes,img_path = sample['image'], sample['density'], sample['boxes'],sample['path']
        if self.dots:
            src_dot = sample['src_dot']
            return {
                'image': self.to_tensor(image),
                'density': torch.from_numpy(density).unsqueeze(0),
                'boxes': boxes,
                'path': img_path,
                'src_dot': torch.from_numpy(src_dot).unsqueeze(0),
                'points': sample['points']
            }
        
        else:
            return {
                'image': self.to_tensor(image),
                'density': torch.from_numpy(density).unsqueeze(0),
                'boxes': boxes,
                'path': img_path,
            }

class Normalize(object):
    def __init__(self, mean, std,dots=False):
        self.normalize = transforms.Normalize(mean = mean, std = std)
        self.dots = dots
    def __call__(self, sample):
        image, density, boxes,img_path = sample['image'], sample['density'], sample['boxes'],sample['path']
        #N = boxes.shape[0]
        
        normalized_image = self.normalize(image)
        #normalized_examples = torch.stack([self.normalize(examples[i]) for i in range(N)])

        if self.dots:
            #image_paste = sample['image_paste']
            #image_paste = self.normalize(image_paste)
            src_dot = sample['src_dot']
            return {
                'image': normalized_image,
                'density': density,
                'boxes': boxes,
                'path': img_path,
                'src_dot': src_dot,
                'points': sample['points']
            }
        
        else:
            return {
                'image': normalized_image,
                'density': density,
                'boxes': boxes,
                'path': img_path,
            }

class CropImage(object):
    """
    Resize the image in a sample to the nearest multiple of 8.
    """
    
    def __init__(self, crop=True,cropSize = 384,dots=False):
        #self.MAX_HW = MAX_HW
        self.crop = crop
        self.dots = dots
        self.cropSize = cropSize

    def __call__(self, sample):
        image, density,img_path = sample['image'], sample['density'],sample['path']
        th,tw = self.cropSize,self.cropSize
        h,w = image.shape[1],image.shape[2]
        x1 = random.randint(0, w - tw)
        y1 = random.randint(0, h - th)  # not used since all images are 384 * Var W  
        x2 = x1 + tw
        points = sample['points']
        boxes = sample['boxes']
        M = boxes.shape[0]
        #print(M,'crop')
        #print(boxes)
        #boxes_old = copy.deepcopy(boxes)
        imageA = image[:,:,x1:x1+tw]
        densityA = density[:,:,x1:x1+tw]          
        for j in range(0,M): # for each provided examplar bounding box
            x1box,y1box = math.ceil(boxes[j,1]) ,math.ceil(boxes[j,2])
            x2box, y2box = math.ceil(boxes[j,3]) ,math.ceil(boxes[j,4]) 
            box_width = x2box - x1box
            box_height = y2box - y1box
            #if not (x1box >= x1 and x2box <= x2):
            if x1box >= x1 and x2box <= x2 and y1box > 0 and y2box <= h:
                x1box = x1box - x1
                x2box = x2box - x1
                y1box = y1box # since cropping is done along X-axis only
                y2box = y2box
                boxes[j,1], boxes[j,2], boxes[j,3], boxes[j,4] = int(x1box), int(y1box), int(x2box), int(y2box)
            else:
                cnter = 0
                for point in points:
                    yc, xc = point[0], point[1]
                    x1box = xc - box_width/2.
                    x2box = xc + (box_width - box_width/2.)
                    y1box = yc - box_height / 2.
                    y2box = yc + (box_height - box_height/2.)
                    if x1box >= x1 and x2box <= x2 and y1box > 0 and y2box <= h:
                        x1box = x1box - x1
                        x2box = x2box - x1
                        y1box = y1box # since cropping is done along X-axis only
                        y2box = y2box
                        boxes[j,1], boxes[j,2], boxes[j,3], boxes[j,4] = int(x1box), int(y1box), int(x2box), int(y2box)
                        break
                    boxes[j,1], boxes[j,2], boxes[j,3], boxes[j,4] = 0, 0, 0, 0
        non_zero_id = torch.nonzero(boxes.sum(dim=-1)).squeeze().tolist()
        if isinstance(non_zero_id, int):
            non_zero_id = [non_zero_id]
        if len(non_zero_id) > 0:
            for j in range(0,M): # replace bounding box outside the crop with one from inside
                if boxes[j].sum() == 0:
                    boxes[j,:] = boxes[non_zero_id[0]]
        else:
            densityA = densityA.fill_(0) # if there's zero exemplar inside the crop, fill GT with 0. Only used for training
            # since testing does not use crops
        if self.dots:
            src_dot = sample['src_dot']
            #image_paste = sample['image_paste']
            src_dotA = src_dot[:,:,x1:x1+tw] 
            #tgt_dotA = tgt_dot[:,:,x1:x1+tw]
            #image_pasteA = image_paste[:,:,x1:x1+tw]
            #tgt_densityA = tgt_density[:,:,x1:x1+tw]
            return {
                'image': imageA,
                'density': densityA,
                'boxes': boxes.unsqueeze(0),
                'path': img_path,
                'src_dot': src_dotA
            }  
        else:
            return {
                'image': imageA, 
                'density': densityA,
                'boxes': boxes.unsqueeze(0),
                'path': img_path}