import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
from collections import OrderedDict

class Resnet50FPN(nn.Module):
    def __init__(self):
        super(Resnet50FPN, self).__init__()
        self.resnet = torchvision.models.resnet50(pretrained=True)
        for param in self.resnet.parameters():
            param.requires_grad = False
        children = list(self.resnet.children())
        self.conv1 = nn.Sequential(*children[:4])
        self.conv2 = children[4]
        self.conv3 = children[5]
        self.conv4 = children[6]
        #self.conv5 = children[7]
    def forward(self, im_data):
        feat = OrderedDict()
        feat_map1 = self.conv1(im_data)
        feat_map1 = self.conv2(feat_map1)
        feat_map3 = self.conv3(feat_map1)
        feat_map4 = self.conv4(feat_map3)
        #feat_map5 = self.conv5(feat_map4)
        #feat['map1'] = feat_map1
        #feat['map2'] = feat_map2
        feat['map3'] = feat_map3
        feat['map4'] = feat_map4
        #feat['map5'] = feat_map5
        return feat

def weights_normal_init(model, dev=0.01):
    if isinstance(model, list):
        for m in model:
            weights_normal_init(m, dev)
    else:
        for m in model.modules():
            if isinstance(m, nn.Conv2d):                
                m.weight.data.normal_(0.0, dev)
                if m.bias is not None:
                    m.bias.data.fill_(0.0)
            elif isinstance(m, nn.Linear):
                m.weight.data.normal_(0.0, dev)



def weights_xavier_init(m):
    if isinstance(m, nn.Conv2d):
        torch.nn.init.xavier_normal_(m.weight, gain=nn.init.calculate_gain('relu'))
        if m.bias is not None:
            torch.nn.init.zeros_(m.bias)




"""

Density Regressor

"""
class CountRegressor(nn.Module):

    def __init__(self, input_channels,pool='mean'):
        super(CountRegressor, self).__init__()
        self.pool = pool
        self.regressor = nn.Sequential(
            nn.Conv2d(input_channels, 196, 7, padding=3),
            nn.InstanceNorm2d(196),
            nn.LeakyReLU(0.2, True),
            nn.UpsamplingBilinear2d(scale_factor=2),
            nn.Conv2d(196, 128, 5, padding=2),
            nn.InstanceNorm2d(128),
            nn.LeakyReLU(0.2, True),
            nn.UpsamplingBilinear2d(scale_factor=2),
            nn.Conv2d(128, 64, 3, padding=1),
            nn.InstanceNorm2d(64),
            nn.LeakyReLU(0.2, True),
            nn.UpsamplingBilinear2d(scale_factor=2),
            nn.Conv2d(64, 32, 1),
            nn.InstanceNorm2d(32),
            nn.LeakyReLU(0.2, True),
            nn.Conv2d(32, 1, 1),
            nn.ReLU(),
        )

    def forward(self, im):
        num_sample =  im.shape[0]
        if num_sample == 1:
            output = self.regressor(im.squeeze(0))
            #print(output2.shape)
            if self.pool == 'mean':
                output = torch.mean(output, dim=(0),keepdim=True)  
                return output
            elif self.pool == 'max':
                output, _ = torch.max(output, 0,keepdim=True)
                return output
        else:
            for i in range(0,num_sample):
                output = self.regressor(im[i])
                if self.pool == 'mean':
                    output = torch.mean(output, dim=(0),keepdim=True)
                elif self.pool == 'max':
                    output, _ = torch.max(output, 0,keepdim=True)
                if i == 0:
                    Output = 1.0 * output
                else:
                    Output = torch.cat((Output,output),dim=0)
            return Output




            
"""

Feature Extractor

"""

def extract_features(feature_model, image, examples):
    
    N, M = image.shape[0], examples.shape[1]
    
    """
    Getting features for the image N * C * H * W
    """
    image_features = feature_model(image)

    """
    Getting features for the examples (N*M) * C * h * w
    """
    examples_features = feature_model(
        examples.view(-1, examples.shape[2], examples.shape[3], examples.shape[4])
    )

    examples_features = examples_features.view(
        N, M, examples_features.shape[1], examples_features.shape[2], examples_features.shape[3])
    
    """
    Convolving example features over image features
    """
    h, w = examples_features.shape[3], examples_features.shape[4]
    
    features = torch.cat([
        F.conv2d(
            F.pad(image_features[i:i+1], ((int(w/2)), int((w-1)/2), int(h/2), int((h-1)/2))),
            examples_features[i]
        ) 
        for i in range(N)
    ])
        
    """
    Averaging and Max Pooling on the M features
    """
    features_avg = torch.mean(features, dim=(1), keepdim=True)    
    features_max, _ = torch.max(features, 1, keepdim=True)
    
    combined = torch.cat([features_avg, features_max], axis = 1)

    return combined

def extract_featuresScaled1(feature_model, image, examples, Example_Scales):
   
    N, M = image.shape[0], examples.shape[1]

    """
    Getting features for the image N * C * H * W
    """
    image_features = feature_model(image)

    """
    Getting features for the examples (N*M) * C * h * w
    """
    examples_features = feature_model(
        examples.view(-1, examples.shape[2], examples.shape[3], examples.shape[4])
    )

    examples_features = examples_features.view(
        N, M, examples_features.shape[1], examples_features.shape[2], examples_features.shape[3])

    """
    Convolving example features over image features
    """
    h, w = examples_features.shape[3], examples_features.shape[4]

    features = torch.cat([
        F.conv2d(
            F.pad(image_features[i:i+1], ((int(w/2)), int((w-1)/2), int(h/2), int((h-1)/2))),
            examples_features[i]
        ) 
        for i in range(N)
    ])

    """
    Averaging and Max Pooling on the M features
    """
    features_avg0 = torch.mean(features, dim=(1), keepdim=True)    
    features_max0, _ = torch.max(features, 1, keepdim=True)
    cnter = 0
    for scale in Example_Scales:
        examples1 = examples.view(-1, examples.shape[2], examples.shape[3], examples.shape[4])
        hscale = int(scale * examples.shape[3])
        wscale = int(scale * examples.shape[4])
        examples1 = F.interpolate(examples1, size=(hscale,wscale),mode='bilinear')
        examples_features = feature_model(examples1)
        examples_features = examples_features.view(
            N, M, examples_features.shape[1], examples_features.shape[2], examples_features.shape[3])

        """
        Convolving example features over image features
        """
        h, w = examples_features.shape[3], examples_features.shape[4]

        features = torch.cat([
            F.conv2d(
                F.pad(image_features[i:i+1], ((int(w/2)), int((w-1)/2), int(h/2), int((h-1)/2))),
                examples_features[i]
            ) 
            for i in range(N)
        ])

        """
        Averaging and Max Pooling on the M features
        """
        features_avg1 = torch.mean(features, dim=(1), keepdim=True)    
        features_max1, _ = torch.max(features, 1, keepdim=True)
        if cnter > 0:
            features_avg = torch.cat((features_avg,features_avg1),1)
            features_max = torch.cat((features_max,features_max1),1)
        else:
            features_avg =torch.cat((features_avg0,features_avg1),1)
            features_max = torch.cat((features_max0,features_max1),1)
        cnter += 1
    #features_avg = torch.mean(features_avg, dim=(1), keepdim=True)    
    #features_max, _ = torch.max(features_max, 1, keepdim=True)
    combined = torch.cat([features_avg, features_max], axis = 1)
    return combined

