import json
import numpy as np
import os
import random
import time
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import sys
from collections import OrderedDict

from os.path import basename, exists, join, splitext
import math
from torch.utils.data import DataLoader
from torchvision import transforms
import copy
import torchvision
from torch.utils.tensorboard import SummaryWriter
from random import randint
from utilsGAN import getGaussianFromDotA,visualizeOutputQuery,closest_dist,apply_gaussian_mask
from utilsDemo import pad_to_size

def FeatureCovariance(A,B):
    '''
    Tensors A and B are of size NDHW
    '''
    A_mean = torch.einsum('ndij->nd',A) / (A.shape[-1] * A.shape[-2])
    B_mean = torch.einsum('ndij->nd',B) / (B.shape[-1] * B.shape[-2])
    A = A - A_mean.view(A_mean.shape[0],A_mean.shape[1],1,1)
    B = B - B_mean.view(B_mean.shape[0],B_mean.shape[1],1,1)
    Cov = torch.einsum('ndij,ndij->',[A,B]) / (torch.numel(A))
    return Cov


def DotProduct(tensor_A,tensor_B,epsilon=1e-6):
    '''
    Computes dot product of two tensors.
    Tensors are normalized, and a small value epsilon is added to the norm for stability 
    '''
    #tensor_A = tensor_A - tensor_A.mean().squeeze()
    #tensor_B = tensor_B - tensor_B.mean().squeeze()
    dotproduct = torch.dot(tensor_A.view(-1)/(tensor_A.norm().item() + epsilon),tensor_B.view(-1)/(tensor_B.norm().item() + epsilon))
    return dotproduct


def extract_featuresWithFeat(feature_model, images, Boxes, feat_map_keys=['map3','map4'], exemplar_scales=[0.9, 1.1],gradientPropagate=False,returnFeat=False,feat_key = 'map4'):
    """
        images: NxCxHxW tensor, where N is the number images in the batch, C is the number of channels
        boxes_ix:  Nx1xMx5 tensor, where M is the number of bounding boxes_ix per images
        return: features for the exemplars of size N * M * (#scales x #maps) * h * w
    """
    Scales = exemplar_scales
    N, M = images.shape[0], Boxes.shape[2]
    #with torch.no_grad(): images_features = feature_model(images)
    if gradientPropagate:
        images_features = feature_model(images)
    else:
        with torch.no_grad():images_features = feature_model(images)        
    #All_feat = list()
    for ix in range(0,N):
        boxes_ix = Boxes[ix].squeeze(0)
        Combined = None
        cnter = 0
        for keys in feat_map_keys:
            im_feats = images_features[keys][ix].unsqueeze(0)
            feat_h = im_feats.shape[-2]
            feat_w = im_feats.shape[-1]

            if keys == 'map1' or keys == 'map2':
                feat_map_scale = 4.0
            elif keys == 'map3':
                feat_map_scale = 8.0
            elif keys == 'map4':
                feat_map_scale = 16.0
            else:
                feat_map_scale = 32.0

            boxes_ix_s = boxes_ix/feat_map_scale

            # boxes_ix_s = torch.round(boxes_ix/feat_map_scale)
            # It is better to extend the box a bit rather than rounding
            boxes_ix_s[:, 1:3] = torch.floor(boxes_ix_s[:, 1:3])
            boxes_ix_s[:, 3:5] = torch.ceil(boxes_ix_s[:, 3:5])

            boxes_ix_s[:, 3:5] = boxes_ix_s[:, 3:5] + 1 # make the end index exclusive for convinient of processing

            # make sure it doesn't go out of bound
            boxes_ix_s[:, 1:3] = torch.clamp_min(boxes_ix_s[:, 1:3], 0)
            boxes_ix_s[:, 3] = torch.clamp_max(boxes_ix_s[:, 3], feat_w)
            boxes_ix_s[:, 4] = torch.clamp_max(boxes_ix_s[:, 4], feat_h)

            box_ws = boxes_ix_s[:, 3] - boxes_ix_s[:, 1]
            box_hs = boxes_ix_s[:, 4] - boxes_ix_s[:, 2]
            max_w = math.ceil(max(exemplar_scales)*max(box_ws))
            max_h = math.ceil(max(exemplar_scales)*max(box_hs))
            
            
            exemplar_feats = torch.zeros(((len(exemplar_scales) + 1)*M, im_feats.shape[1], max_h, max_w), device=im_feats.device)

            cnt = 0
           
            for j in range(0,M): # for each provided examplar bounding box
                x1, y1, x2, y2 = int(boxes_ix_s[j, 1]), int(boxes_ix_s[j, 2]), int(boxes_ix_s[j, 3]), int(boxes_ix_s[j, 4])
                ori_feat = im_feats[:, :, y1:y2, x1:x2]
                h, w = ori_feat.shape[-2], ori_feat.shape[-1]                
                #ori_feat = Ori_feat[j]
                #print(exemplar_feats.shape,ori_feat.shape)
                h, w = ori_feat.shape[-2], ori_feat.shape[-1]
                exemplar_feats[cnt] = pad_to_size(ori_feat, max_h, max_w)
                cnt += 1

                for scale in Scales:
                    scale_h = math.ceil(scale * h)
                    scale_w = math.ceil(scale * w)
                    scale_feat = F.interpolate(ori_feat, size=(scale_h, scale_w), mode='bilinear')
                    exemplar_feats[cnt] = pad_to_size(scale_feat, max_h, max_w)
                    cnt += 1            
            h, w = exemplar_feats.shape[-2], exemplar_feats.shape[-1]
            #examples_features = examples_features 
            features =    F.conv2d(
                    F.pad(im_feats, ((int(w/2)), int((w-1)/2), int(h/2), int((h-1)/2))),
                    exemplar_feats
                )
            #print(M,image_features.shape,features.shape,examples_features.shape,h,w)
            #features_avg = torch.mean(features, dim=(1), keepdim=True)    
            #features_max, _ = torch.max(features, 1, keepdim=True)
            #nos_examples = int(examples_features.shape[0]/M)
            #combined = torch.reshape(features, (M, len(Scales) + 1, features.shape[-2], features.shape[-1]))
            combined = features.view(M, len(exemplar_scales) + 1, features.shape[-2], features.shape[-1])
                    
                    
            if cnter == 0:
                Combined = 1.0 * combined
            else:
                if Combined.shape[2] != combined.shape[2] or Combined.shape[3] != combined.shape[3]:
                    combined = F.interpolate(combined, size=(Combined.shape[2],Combined.shape[3]),mode='bilinear')
                Combined = torch.cat((Combined,combined),dim=1)
            cnter += 1
        if ix == 0:
            All_feat = 1.0 * Combined.unsqueeze(0)
        else:
            All_feat = torch.cat((All_feat,Combined.unsqueeze(0)),dim=0)
    if not returnFeat:
        return All_feat
    else:
        return All_feat,images_features[feat_key]

def extract_featuresManifoldMixup(feature_model, images, Boxes, feat_map_keys=['map3','map4'], exemplar_scales=[0.9, 1.1],gradientPropagate=False,alpha = 1.0):
    """
        images: NxCxHxW tensor, where N is the number images in the batch, C is the number of channels
        boxes_ix:  Nx1xMx5 tensor, where M is the number of bounding boxes_ix per images
        return: features for the exemplars of size N * M * (#scales x #maps) * h * w
    """
    Scales = exemplar_scales
    N, M = images.shape[0], Boxes.shape[2]
    assert N > 1
    if alpha > 0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1
    index = torch.randperm(N).cuda()
    #with torch.no_grad(): images_features = feature_model(images)
    if gradientPropagate:
        images_features = feature_model(images)
    else:
        with torch.no_grad():images_features = feature_model(images)        
    #All_feat = list()
    for ix in range(0,N):
        boxes_ix = Boxes[ix].squeeze(0)
        boxes_iy = Boxes[index[ix]].squeeze(0)
        Combined = None
        cnter = 0
        for keys in feat_map_keys:
            im_feats = images_features[keys][ix].unsqueeze(0)
            im_featsy = images_features[keys][index[ix]].unsqueeze(0)
            feat_h = im_feats.shape[-2]
            feat_w = im_feats.shape[-1]

            if keys == 'map1' or keys == 'map2':
                feat_map_scale = 4.0
            elif keys == 'map3':
                feat_map_scale = 8.0
            elif keys == 'map4':
                feat_map_scale = 16.0
            else:
                feat_map_scale = 32.0

            boxes_ix_s = boxes_ix/feat_map_scale

            # boxes_ix_s = torch.round(boxes_ix/feat_map_scale)
            # It is better to extend the box a bit rather than rounding
            boxes_ix_s[:, 1:3] = torch.floor(boxes_ix_s[:, 1:3])
            boxes_ix_s[:, 3:5] = torch.ceil(boxes_ix_s[:, 3:5])

            boxes_ix_s[:, 3:5] = boxes_ix_s[:, 3:5] + 1 # make the end index exclusive for convinient of processing

            # make sure it doesn't go out of bound
            boxes_ix_s[:, 1:3] = torch.clamp_min(boxes_ix_s[:, 1:3], 0)
            boxes_ix_s[:, 3] = torch.clamp_max(boxes_ix_s[:, 3], feat_w)
            boxes_ix_s[:, 4] = torch.clamp_max(boxes_ix_s[:, 4], feat_h)

            box_ws = boxes_ix_s[:, 3] - boxes_ix_s[:, 1]
            box_hs = boxes_ix_s[:, 4] - boxes_ix_s[:, 2]
            max_w = math.ceil(max(exemplar_scales)*max(box_ws))
            max_h = math.ceil(max(exemplar_scales)*max(box_hs))
            

            boxes_ix_sy = boxes_iy/feat_map_scale

            # boxes_ix_s = torch.round(boxes_ix/feat_map_scale)
            # It is better to extend the box a bit rather than rounding
            boxes_ix_sy[:, 1:3] = torch.floor(boxes_ix_sy[:, 1:3])
            boxes_ix_sy[:, 3:5] = torch.ceil(boxes_ix_sy[:, 3:5])

            boxes_ix_sy[:, 3:5] = boxes_ix_sy[:, 3:5] + 1 # make the end index exclusive for convinient of processing

            # make sure it doesn't go out of bound
            boxes_ix_sy[:, 1:3] = torch.clamp_min(boxes_ix_sy[:, 1:3], 0)
            boxes_ix_sy[:, 3] = torch.clamp_max(boxes_ix_sy[:, 3], feat_w)
            boxes_ix_sy[:, 4] = torch.clamp_max(boxes_ix_sy[:, 4], feat_h)

            box_wsy = boxes_ix_sy[:, 3] - boxes_ix_sy[:, 1]
            box_hsy = boxes_ix_sy[:, 4] - boxes_ix_sy[:, 2]
            max_wy = math.ceil(max(exemplar_scales)*max(box_wsy))
            max_hy = math.ceil(max(exemplar_scales)*max(box_hsy))
            
            max_hjoint = max(max_h,max_hy)
            max_wjoint = max(max_w,max_wy)
            exemplar_feats = torch.zeros(((len(exemplar_scales) + 1)*M, im_feats.shape[1], max_hjoint, max_wjoint), device=im_feats.device)

            cnt = 0
           
            for j in range(0,M): # for each provided examplar bounding box
                x1, y1, x2, y2 = int(boxes_ix_s[j, 1]), int(boxes_ix_s[j, 2]), int(boxes_ix_s[j, 3]), int(boxes_ix_s[j, 4])
                ori_feat = im_feats[:, :, y1:y2, x1:x2]
                h, w = ori_feat.shape[-2], ori_feat.shape[-1]                
                #ori_feat = Ori_feat[j]
                #print(exemplar_feats.shape,ori_feat.shape)
                h, w = ori_feat.shape[-2], ori_feat.shape[-1]
                exemplar_feats[cnt] = pad_to_size(ori_feat, max_hjoint, max_wjoint)
                cnt += 1

                for scale in Scales:
                    scale_h = math.ceil(scale * h)
                    scale_w = math.ceil(scale * w)
                    scale_feat = F.interpolate(ori_feat, size=(scale_h, scale_w), mode='bilinear')
                    exemplar_feats[cnt] = pad_to_size(scale_feat, max_hjoint, max_wjoint)
                    cnt += 1 

            exemplar_featsy = torch.zeros(((len(exemplar_scales) + 1)*M, im_featsy.shape[1], max_hjoint, max_wjoint), device=im_feats.device)

            cnt = 0
           
            for j in range(0,M): # for each provided examplar bounding box
                x1, y1, x2, y2 = int(boxes_ix_sy[j, 1]), int(boxes_ix_sy[j, 2]), int(boxes_ix_sy[j, 3]), int(boxes_ix_sy[j, 4])
                ori_feat = im_featsy[:, :, y1:y2, x1:x2]
                h, w = ori_feat.shape[-2], ori_feat.shape[-1]                
                #ori_feat = Ori_feat[j]
                #print(exemplar_feats.shape,ori_feat.shape)
                h, w = ori_feat.shape[-2], ori_feat.shape[-1]
                exemplar_featsy[cnt] = pad_to_size(ori_feat, max_hjoint, max_wjoint)
                cnt += 1

                for scale in Scales:
                    scale_h = math.ceil(scale * h)
                    scale_w = math.ceil(scale * w)
                    scale_feat = F.interpolate(ori_feat, size=(scale_h, scale_w), mode='bilinear')
                    exemplar_featsy[cnt] = pad_to_size(scale_feat, max_hjoint, max_wjoint)
                    cnt += 1 
            h, w = exemplar_feats.shape[-2], exemplar_feats.shape[-1]
            #examples_features = examples_features 
            exemplar_feats = lam * exemplar_feats + (1 - lam) * exemplar_featsy
            im_feats = lam * im_feats + (1 - lam) * im_featsy
            features =    F.conv2d(
                    F.pad(im_feats, ((int(w/2)), int((w-1)/2), int(h/2), int((h-1)/2))),
                    exemplar_feats
                )
            #print(M,image_features.shape,features.shape,examples_features.shape,h,w)
            #features_avg = torch.mean(features, dim=(1), keepdim=True)    
            #features_max, _ = torch.max(features, 1, keepdim=True)
            #nos_examples = int(examples_features.shape[0]/M)
            #combined = torch.reshape(features, (M, len(Scales) + 1, features.shape[-2], features.shape[-1]))
            combined = features.view(M, len(exemplar_scales) + 1, features.shape[-2], features.shape[-1])
                    
                    
            if cnter == 0:
                Combined = 1.0 * combined
            else:
                if Combined.shape[2] != combined.shape[2] or Combined.shape[3] != combined.shape[3]:
                    combined = F.interpolate(combined, size=(Combined.shape[2],Combined.shape[3]),mode='bilinear')
                Combined = torch.cat((Combined,combined),dim=1)
            cnter += 1
        if ix == 0:
            All_feat = 1.0 * Combined.unsqueeze(0)
        else:
            All_feat = torch.cat((All_feat,Combined.unsqueeze(0)),dim=0)
    return All_feat,index,lam

def extract_featuresInputMixup(feature_model, images, Boxes, feat_map_keys=['map3','map4'], exemplar_scales=[0.9, 1.1],gradientPropagate=False,alpha = 1.0):
    """
        images: NxCxHxW tensor, where N is the number images in the batch, C is the number of channels
        boxes_ix:  Nx1xMx5 tensor, where M is the number of bounding boxes_ix per images
        return: features for the exemplars of size N * M * (#scales x #maps) * h * w
    """
    Scales = exemplar_scales
    N, M = images.shape[0], Boxes.shape[2]
    assert N > 1
    if alpha > 0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1
    index = torch.randperm(N).cuda()
    #with torch.no_grad(): images_features = feature_model(images)
    images_mixup = images[index,:]
    images = lam * images + (1 - lam) * images_mixup
    
    if gradientPropagate:
        images_features = feature_model(images)
    else:
        with torch.no_grad():images_features = feature_model(images)        
    #All_feat = list()
    for ix in range(0,N):
        boxes_ix = Boxes[ix].squeeze(0)
        boxes_iy = Boxes[index[ix]].squeeze(0)
        boxes_iy = torch.cat((boxes_ix,boxes_iy),dim = 0)
        Combined = None
        cnter = 0
        for keys in feat_map_keys:
            im_feats = images_features[keys][ix].unsqueeze(0)
            #im_featsy = images_features[keys][index[ix]].unsqueeze(0)
            feat_h = im_feats.shape[-2]
            feat_w = im_feats.shape[-1]

            if keys == 'map1' or keys == 'map2':
                feat_map_scale = 4.0
            elif keys == 'map3':
                feat_map_scale = 8.0
            elif keys == 'map4':
                feat_map_scale = 16.0
            else:
                feat_map_scale = 32.0

            boxes_iy_s = boxes_iy/feat_map_scale

            # boxes_ix_s = torch.round(boxes_ix/feat_map_scale)
            # It is better to extend the box a bit rather than rounding
            boxes_iy_s[:, 1:3] = torch.floor(boxes_iy_s[:, 1:3])
            boxes_iy_s[:, 3:5] = torch.ceil(boxes_iy_s[:, 3:5])

            boxes_iy_s[:, 3:5] = boxes_iy_s[:, 3:5] + 1 # make the end index exclusive for convinient of processing

            # make sure it doesn't go out of bound
            boxes_iy_s[:, 1:3] = torch.clamp_min(boxes_iy_s[:, 1:3], 0)
            boxes_iy_s[:, 3] = torch.clamp_max(boxes_iy_s[:, 3], feat_w)
            boxes_iy_s[:, 4] = torch.clamp_max(boxes_iy_s[:, 4], feat_h)

            box_ws = boxes_iy_s[:, 3] - boxes_iy_s[:, 1]
            box_hs = boxes_iy_s[:, 4] - boxes_iy_s[:, 2]
            max_w = math.ceil(max(exemplar_scales)*max(box_ws))
            max_h = math.ceil(max(exemplar_scales)*max(box_hs))
            


            exemplar_feats = torch.zeros(((len(exemplar_scales) + 1)*2*M, im_feats.shape[1], max_h, max_w), device=im_feats.device)

            cnt = 0
           
            for j in range(0,2*M): # for each provided examplar bounding box
                x1, y1, x2, y2 = int(boxes_iy_s[j, 1]), int(boxes_iy_s[j, 2]), int(boxes_iy_s[j, 3]), int(boxes_iy_s[j, 4])
                ori_feat = im_feats[:, :, y1:y2, x1:x2]
                h, w = ori_feat.shape[-2], ori_feat.shape[-1]                
                #ori_feat = Ori_feat[j]
                #print(exemplar_feats.shape,ori_feat.shape)
                h, w = ori_feat.shape[-2], ori_feat.shape[-1]
                exemplar_feats[cnt] = pad_to_size(ori_feat, max_h, max_w)
                cnt += 1

                for scale in Scales:
                    scale_h = math.ceil(scale * h)
                    scale_w = math.ceil(scale * w)
                    scale_feat = F.interpolate(ori_feat, size=(scale_h, scale_w), mode='bilinear')
                    exemplar_feats[cnt] = pad_to_size(scale_feat, max_h, max_w)
                    cnt += 1 

            h, w = exemplar_feats.shape[-2], exemplar_feats.shape[-1]
            features =    F.conv2d(
                    F.pad(im_feats, ((int(w/2)), int((w-1)/2), int(h/2), int((h-1)/2))),
                    exemplar_feats
                )
            #print(M,image_features.shape,features.shape,examples_features.shape,h,w)
            #features_avg = torch.mean(features, dim=(1), keepdim=True)    
            #features_max, _ = torch.max(features, 1, keepdim=True)
            #nos_examples = int(examples_features.shape[0]/M)
            #combined = torch.reshape(features, (M, len(Scales) + 1, features.shape[-2], features.shape[-1]))
            combined = features.view(2*M, len(exemplar_scales) + 1, features.shape[-2], features.shape[-1])
                    
                    
            if cnter == 0:
                Combined = 1.0 * combined
            else:
                if Combined.shape[2] != combined.shape[2] or Combined.shape[3] != combined.shape[3]:
                    combined = F.interpolate(combined, size=(Combined.shape[2],Combined.shape[3]),mode='bilinear')
                Combined = torch.cat((Combined,combined),dim=1)
            cnter += 1
        if ix == 0:
            All_feat = 1.0 * Combined.unsqueeze(0)
        else:
            All_feat = torch.cat((All_feat,Combined.unsqueeze(0)),dim=0)
    return All_feat,index,lam


def extract_features(feature_model, images, Boxes, feat_map_keys=['map3','map4'], exemplar_scales=[0.9, 1.1],gradientPropagate=False):
    """
        images: NxCxHxW tensor, where N is the number images in the batch, C is the number of channels
        boxes_ix:  Nx1xMx5 tensor, where M is the number of bounding boxes_ix per images
        return: features for the exemplars of size N * M * (#scales x #maps) * h * w
    """
    Scales = exemplar_scales
    N, M = images.shape[0], Boxes.shape[2]
    #with torch.no_grad(): images_features = feature_model(images)
    if gradientPropagate:
        images_features = feature_model(images)
    else:
        with torch.no_grad():images_features = feature_model(images)        
    #All_feat = list()
    for ix in range(0,N):
        boxes_ix = Boxes[ix].squeeze(0)
        Combined = None
        cnter = 0
        for keys in feat_map_keys:
            im_feats = images_features[keys][ix].unsqueeze(0)
            feat_h = im_feats.shape[-2]
            feat_w = im_feats.shape[-1]

            if keys == 'map1' or keys == 'map2':
                feat_map_scale = 4.0
            elif keys == 'map3':
                feat_map_scale = 8.0
            elif keys == 'map4':
                feat_map_scale = 16.0
            else:
                feat_map_scale = 32.0

            boxes_ix_s = boxes_ix/feat_map_scale

            # boxes_ix_s = torch.round(boxes_ix/feat_map_scale)
            # It is better to extend the box a bit rather than rounding
            boxes_ix_s[:, 1:3] = torch.floor(boxes_ix_s[:, 1:3])
            boxes_ix_s[:, 3:5] = torch.ceil(boxes_ix_s[:, 3:5])

            boxes_ix_s[:, 3:5] = boxes_ix_s[:, 3:5] + 1 # make the end index exclusive for convinient of processing

            # make sure it doesn't go out of bound
            boxes_ix_s[:, 1:3] = torch.clamp_min(boxes_ix_s[:, 1:3], 0)
            boxes_ix_s[:, 3] = torch.clamp_max(boxes_ix_s[:, 3], feat_w)
            boxes_ix_s[:, 4] = torch.clamp_max(boxes_ix_s[:, 4], feat_h)

            box_ws = boxes_ix_s[:, 3] - boxes_ix_s[:, 1]
            box_hs = boxes_ix_s[:, 4] - boxes_ix_s[:, 2]
            max_w = math.ceil(max(exemplar_scales)*max(box_ws))
            max_h = math.ceil(max(exemplar_scales)*max(box_hs))
            
            
            exemplar_feats = torch.zeros(((len(exemplar_scales) + 1)*M, im_feats.shape[1], max_h, max_w), device=im_feats.device)

            cnt = 0
           
            for j in range(0,M): # for each provided examplar bounding box
                x1, y1, x2, y2 = int(boxes_ix_s[j, 1]), int(boxes_ix_s[j, 2]), int(boxes_ix_s[j, 3]), int(boxes_ix_s[j, 4])
                ori_feat = im_feats[:, :, y1:y2, x1:x2]
                h, w = ori_feat.shape[-2], ori_feat.shape[-1]                
                #ori_feat = Ori_feat[j]
                #print(exemplar_feats.shape,ori_feat.shape)
                h, w = ori_feat.shape[-2], ori_feat.shape[-1]
                exemplar_feats[cnt] = pad_to_size(ori_feat, max_h, max_w)
                cnt += 1

                for scale in Scales:
                    scale_h = math.ceil(scale * h)
                    scale_w = math.ceil(scale * w)
                    scale_feat = F.interpolate(ori_feat, size=(scale_h, scale_w), mode='bilinear')
                    exemplar_feats[cnt] = pad_to_size(scale_feat, max_h, max_w)
                    cnt += 1            
            h, w = exemplar_feats.shape[-2], exemplar_feats.shape[-1]
            #examples_features = examples_features 
            features =    F.conv2d(
                    F.pad(im_feats, ((int(w/2)), int((w-1)/2), int(h/2), int((h-1)/2))),
                    exemplar_feats
                )
            #print(M,image_features.shape,features.shape,examples_features.shape,h,w)
            #features_avg = torch.mean(features, dim=(1), keepdim=True)    
            #features_max, _ = torch.max(features, 1, keepdim=True)
            #nos_examples = int(examples_features.shape[0]/M)
            #combined = torch.reshape(features, (M, len(Scales) + 1, features.shape[-2], features.shape[-1]))
            combined = features.view(M, len(exemplar_scales) + 1, features.shape[-2], features.shape[-1])
                    
                    
            if cnter == 0:
                Combined = 1.0 * combined
            else:
                if Combined.shape[2] != combined.shape[2] or Combined.shape[3] != combined.shape[3]:
                    combined = F.interpolate(combined, size=(Combined.shape[2],Combined.shape[3]),mode='bilinear')
                Combined = torch.cat((Combined,combined),dim=1)
            cnter += 1
        if ix == 0:
            All_feat = 1.0 * Combined.unsqueeze(0)
        else:
            All_feat = torch.cat((All_feat,Combined.unsqueeze(0)),dim=0)
    return All_feat




def get_Mask(src_dot,boxes):
    boxes = boxes.squeeze()
    Mask = torch.zeros_like(src_dot)
    Scaling = 1.0
    y1,x1,y2,x2 = math.ceil(boxes[0,1]/Scaling),math.ceil(boxes[0,2]/Scaling),math.ceil(boxes[0,3]/Scaling),math.ceil(boxes[0,4]/Scaling)
    boxes1 = 1. * boxes
    X = 1 * (1 + x2 - x1)
    Y = 1 * (1 + y2 - y1)
    #print(X,Y)
    for j in range(0,boxes.shape[0]):
        #y11,x11,y21,x21 = math.ceil(boxes[1]/Scaling),math.ceil(boxes[2]/Scaling),math.ceil(boxes[3]/Scaling),math.ceil(boxes[4]/Scaling)
        #boxes1[1],boxes1[2],boxes1[3],boxes1[4] = boxes[2],boxes[1],boxes[4],boxes[3]        
        y11,x11,y21,x21 = math.ceil(boxes[j,1]/Scaling),math.ceil(boxes[j,2]/Scaling),math.ceil(boxes[j,3]/Scaling),math.ceil(boxes[j,4]/Scaling)
        boxes1[j,1],boxes1[j,2],boxes1[j,3],boxes1[j,4] = boxes[j,2],boxes[j,1],boxes[j,4],boxes[j,3]                
        if x21 - x11 > x2 - x1:
            x2 = 1 * x21
            x1 = 1 * x11
            X = 1 * (x2 - x1) + 1
        if y21 - y11 > y2 - y1:
            y2 = 1 * y21
            y1 = 1* y11
            Y = 1 * (y2 - y1) + 1
        #print(X,Y)
    #print(X,Y)
    H,W =src_dot.shape[-2],src_dot.shape[-1]
    NewLocs = src_dot.squeeze()
    indices=torch.nonzero(NewLocs)
    #print(Mask.sum(),Mask.shape)
    for i in range(0,indices.shape[0]):
        z=[0,0]
        z[0],z[1] = indices[i][0],indices[i][1]
        y1,x1 = int(max(0,z[0].item()-X/2)),int(max(0,z[1].item()-Y/2))
        y2,x2 = int(min(H,z[0].item()+X/2)),int(min(W,z[1].item()+Y/2))
        Mask[0,0,y1:y2,x1:x2] = 1
        #print(x1,x2,y1,y2,H,W,X,Y)
    #print(Mask.sum())
    #print(indices)
    return Mask


# In[11]:


def getGaussianFromDotB(tgt_dot,src_dot,src_density):
    indices_y,indices_x = np.nonzero(src_dot)
    points = list()
    for i in range(0,len(indices_x)):
        points.append([indices_y[i],indices_x[i]])
    #print(indices)
    d = [closest_dist(p, points)[1] for p in points]
    mean_r = np.mean(d) / 2.0

    r_x = max(15, min(60, int(round(mean_r))))
    r_y = max(15, min(60, int(round(mean_r))))
    #print(r_x,r_y)
    H,W = src_dot.shape
    density_map = 1.0 * src_density

    assert r_x == r_y, "Code does not support generation of ellipse"

    sd = round(r_x / 4.0)
    indices_y,indices_x = np.nonzero(abs(src_dot-tgt_dot))
    points = list()
    for i in range(0,len(indices_x)):
        points.append([indices_y[i],indices_x[i]])
    for point in points:
        y,x = point

        r1 = int(round(max(0, y - r_y)))
        c1 = int(round(max(0, x - r_x)))

        r2 = int(round(min(H-1, y + r_y)))
        c2 = int(round(min(W-1, x + r_x)))
        #print(r1,r2,c1,c2,r2-r1+1, c2-c1+1)
        density_map = apply_gaussian_mask(density_map, r1, c1, r2, c2, sd)
    return density_map


# In[12]:


def getGaussianFromDotC(tgt_dot,src_dot,src_density):
    indices_y,indices_x = np.nonzero(src_dot)
    points = list()
    for i in range(0,len(indices_x)):
        points.append([indices_y[i],indices_x[i]])
    #print(indices)
    d = [closest_dist(p, points)[1] for p in points]
    mean_r = np.mean(d) / 2.0

    r_x = max(15, min(60, int(round(mean_r))))
    r_y = max(15, min(60, int(round(mean_r))))
    #print(r_x,r_y)
    H,W = src_dot.shape
    density_map = np.float32(np.zeros((H,W)))#1.0 * src_density

    assert r_x == r_y, "Code does not support generation of ellipse"

    sd = round(r_x / 4.0)
    indices_y,indices_x = np.nonzero(abs(tgt_dot))
    points = list()
    for i in range(0,len(indices_x)):
        points.append([indices_y[i],indices_x[i]])
    for point in points:
        y,x = point

        r1 = int(round(max(0, y - r_y)))
        c1 = int(round(max(0, x - r_x)))

        r2 = int(round(min(H-1, y + r_y)))
        c2 = int(round(min(W-1, x + r_x)))
        #print(r1,r2,c1,c2,r2-r1+1, c2-c1+1)
        density_map = apply_gaussian_mask(density_map, r1, c1, r2, c2, sd)
    return density_map


# ### Loss

# In[13]:

def weighted_L2_loss(Y, target, weights):
    out = (Y - target) ** 2
    out = out * weights.expand_as(out)
    # expand_as because weights are prob not defined for mini-batch
    out = out.mean() # or sum over whatever dimensions
    return out

def weighted_L1_loss(Y, target, weights):
    out = torch.abs(Y - target)
    out = out * weights.expand_as(out)
    # expand_as because weights are prob not defined for mini-batch
    out = out.mean() # or sum over whatever dimensions
    return out

def weighted_center_loss(pred_mask, tgt_density, Mask0):
    Weight = tgt_density * Mask0
    out = torch.abs(pred_mask - Mask0)
    out = out * Weight.expand_as(out)
    # expand_as because weights are prob not defined for mini-batch
    out = out.sum()/Mask0.sum().detach().item() # or sum over whatever dimensions
    return out    



##################   VISUALIZATION UTILS #######################################################
################################################################################################

import decimal
import matplotlib.pyplot as plt
# create a new context for this task
ctx = decimal.Context()

# 20 digits should be enough for everyone :D
ctx.prec = 2

def float_to_str(f):
    """
    Convert the given float to a string,
    without resorting to scientific notation
    """
    d1 = ctx.create_decimal(repr(f))
    return format(d1, 'f')
def visualizeOutputQuery(input_, output,boxes,guided=False, use_gpu=False,
              figsize=(20, 6), cmap='viridis', alpha=.5,
              return_output=False,Cnter=0):

    # Setup subplots
    img1 = format_for_plotting(denormalize(input_))
    img1 = AddBoxes(img1,boxes,Query=None)
    cnt = output.sum().item()
    output = format_for_plotting(standardize_and_clip(output))
    #output = AddBoxesDensity(output,boxes,Query=None)
    subplots = [
        # (title, [(img1, cmap, alpha), (image2, cmap, alpha)])
        ('Input image no. ' + str(int(Cnter)),
         [(img1, None, None)]),
        ('Predicted count '+ str(int(cnt)),
         [(output,
          None,
          None)])    
    ]

    fig = plt.figure(figsize=figsize)

    for i, (title, images) in enumerate(subplots):
        ax = fig.add_subplot(1, len(subplots), i + 1)
        ax.set_axis_off()
        ax.set_title(title)

        for image, cmap, alpha in images:
            ax.imshow(image, cmap=cmap, alpha=alpha)
            
def visualizeExemplar(input_,boxes,guided=False, use_gpu=False,
              figsize=(20, 6), cmap='viridis', alpha=.5,
              return_output=False,Cnter=0):

    # Setup subplots
    img1 = format_for_plotting(denormalize(input_))
    img1 = AddBoxes(img1,boxes,Query=None)
    #cnt = output.sum().item()
    #output = format_for_plotting(standardize_and_clip(output))
    #output = AddBoxesDensity(output,boxes,Query=None)
    subplots = [
        # (title, [(img1, cmap, alpha), (image2, cmap, alpha)])
        ('Input image with Exemplar',
         [(img1, None, None)])]

    fig = plt.figure(figsize=figsize)

    for i, (title, images) in enumerate(subplots):
        ax = fig.add_subplot(1, len(subplots), i + 1)
        ax.set_axis_off()
        ax.set_title(title)

        for image, cmap, alpha in images:
            ax.imshow(image, cmap=cmap, alpha=alpha)
    

def denormalize(tensor):
    """Reverses the normalisation on a tensor.
    Performs a reverse operation on a tensor, so the pixel value range is
    between 0 and 1. Useful for when plotting a tensor into an image.
    Normalisation: (image - mean) / std
    Denormalisation: image * std + mean
    Args:
        tensor (torch.Tensor, dtype=torch.float32): Normalized image tensor
    Shape:
        Input: :math:`(N, C, H, W)`
        Output: :math:`(N, C, H, W)` (same shape as input)
    Return:
        torch.Tensor (torch.float32): Demornalised image tensor with pixel
            values between [0, 1]
    Note:
        Symbols used to describe dimensions:
            - N: number of images in a batch
            - C: number of channels
            - H: height of the image
            - W: width of the image
    """

    means = [0.485, 0.456, 0.406]
    stds = [0.229, 0.224, 0.225]

    denormalized = tensor.clone()

    for channel, mean, std in zip(denormalized[0], means, stds):
        channel.mul_(std).add_(mean)

    return denormalized




def standardize_and_clip(tensor, min_value=0.0, max_value=1.0,
                         saturation=0.1, brightness=0.5):

    """Standardizes and clips input tensor.
    Standardizes the input tensor (mean = 0.0, std = 1.0). The color saturation
    and brightness are adjusted, before tensor values are clipped to min/max
    (default: 0.0/1.0).
    Args:
        tensor (torch.Tensor):
        min_value (float, optional, default=0.0)
        max_value (float, optional, default=1.0)
        saturation (float, optional, default=0.1)
        brightness (float, optional, default=0.5)
    Shape:
        Input: :math:`(C, H, W)`
        Output: Same as the input
    Return:
        torch.Tensor (torch.float32): Normalised tensor with values between
            [min_value, max_value]
    """

    tensor = tensor.detach().cpu()

    mean = tensor.mean()
    std = tensor.std()

    if std == 0:
        std += 1e-7

    standardized = tensor.sub(mean).div(std).mul(saturation)
    clipped = standardized.add(brightness).clamp(min_value, max_value)

    return clipped

def format_for_plotting(tensor):
    """Formats the shape of tensor for plotting.
    Tensors typically have a shape of :math:`(N, C, H, W)` or :math:`(C, H, W)`
    which is not suitable for plotting as images. This function formats an
    input tensor :math:`(H, W, C)` for RGB and :math:`(H, W)` for mono-channel
    data.
    Args:
        tensor (torch.Tensor, torch.float32): Image tensor
    Shape:
        Input: :math:`(N, C, H, W)` or :math:`(C, H, W)`
        Output: :math:`(H, W, C)` or :math:`(H, W)`, respectively
    Return:
        torch.Tensor (torch.float32): Formatted image tensor (detached)
    Note:
        Symbols used to describe dimensions:
            - N: number of images in a batch
            - C: number of channels
            - H: height of the image
            - W: width of the image
    """

    has_batch_dimension = len(tensor.shape) == 4
    formatted = tensor.clone()

    if has_batch_dimension:
        formatted = tensor.squeeze(0)

    if formatted.shape[0] == 1:
        return formatted.squeeze(0).detach()
    else:
        return formatted.permute(1, 2, 0).detach()
    
def AddBoxes(img0,boxes,Query=None):
    #boxes = boxes.squeeze()
    #Query = Query.squeeze()
    #img1 = 1.0 * img0
    img = 1.0 * img0
    if boxes.shape[1] > 1:
        boxes = boxes.squeeze()
        for i in range(0,boxes.shape[0]):
            x1,y1,x2,y2 = int(boxes[i,2].item()),int(boxes[i,1].item()),int(boxes[i,4].item()),int(boxes[i,3].item())
            img[x1:x2,y1-2:y1+2,1] = 1.
            img[x1-2:x1+2,y1:y2,1] = 1.
            img[x1:x2,y2-2:y2+2,1] = 1.
            img[x2-2:x2+2,y1:y2,1] = 1.
    else:
        boxes = boxes.squeeze()
        #for i in range(0,boxes.shape[0]):
        x1,y1,x2,y2 = int(boxes[2].item()),int(boxes[1].item()),int(boxes[4].item()),int(boxes[3].item())
        img[x1:x2,y1-2:y1+2,1] = 1.
        img[x1-2:x1+2,y1:y2,1] = 1.
        img[x1:x2,y2-2:y2+2,1] = 1.
        img[x2-2:x2+2,y1:y2,1] = 1. 
    if Query is not None:
        if Query.shape[1] > 1:
            Query = Query.squeeze()
            for i in range(0,Query.shape[0]):
                x1,y1,x2,y2 = int(Query[i,2].item()),int(Query[i,1].item()),int(Query[i,4].item()),int(Query[i,3].item())
                img[x1:x2,y1-2:y1+2,0] = 1.
                img[x1-2:x1+2,y1:y2,0] = 1.
                img[x1:x2,y2-2:y2+2,0] = 1.
                img[x2-2:x2+2,y1:y2,0] = 1.
        else:
            Query = Query.squeeze()
            x1,y1,x2,y2 = int(Query[2].item()),int(Query[1].item()),int(Query[4].item()),int(Query[3].item())
            img[x1:x2,y1-2:y1+2,0] = 1.
            img[x1-2:x1+2,y1:y2,0] = 1.
            img[x1:x2,y2-2:y2+2,0] = 1.
            img[x2-2:x2+2,y1:y2,0] = 1.        
    return img

